from __future__ import print_function
import sys
import libvirt
import threading
import time
from flask import Flask
app = Flask(__name__)
from flask import render_template

started = False
threads = []
usage_lock = threading.Lock()
usage_count = dict()
total_window = 25
unread = list()
read = list()

def master():
    threshold = 90.0 	#percent
    window = 8
    global usage_lock
    global usage_count
    assert window <= total_window
    
    while True:
        with usage_lock:
            u = usage_count.copy()
        for key in u:
            if key == "loadbalancer":
                continue
            avg = 0.0
            for i in range(-1,-1-window,-1):
                avg += u[key][i]
            avg /= window
            if avg > threshold:
                conn = libvirt.open('qemu:///system')
                dom = conn.lookupByName("server2")
                state, reason = dom.state()
                if state == libvirt.VIR_DOMAIN_RUNNING:
                    continue
                unread.append("Overload Detected! Spawning new VM.")
                if dom.create() < 0:
                    print('Can not boot guest domain.', file=sys.stderr)
                    exit(1)
                conn.close()
                time.sleep(60)              

def calculate_usage(dom):
    global usage_count
    old_time = 0
    delay = 1 # seconds
    while started:
        state, reason = dom.state()
        if state != libvirt.VIR_DOMAIN_RUNNING:
            if dom.name() not in usage_count:
                usage_count[dom.name()] = [0] * total_window
            else:
                usage_count[dom.name()].pop(0)
                usage_count[dom.name()].append(0)
            time.sleep(delay)
            continue
        stats = dom.getCPUStats(True)
        new_time = stats[0]['cpu_time']
        if old_time>0:
            usage = 100 * (new_time - old_time) / (delay * 1000 * 1000 * 1000)
            usage = round(usage, 2)
            with usage_lock:
                if dom.name() not in usage_count:
                    usage_count[dom.name()] = [0] * total_window
                else:
                    usage_count[dom.name()].pop(0)
                    usage_count[dom.name()].append(usage)
        old_time = new_time
        time.sleep(delay)

@app.route('/status')
def getStatus():
    global usage_count
    return usage_count

def start_monitoring():
    global started
    if started:
        return
    started = True
    conn = libvirt.open('qemu:///system')
    if conn == None:
        print('Failed to open connection to qemu:///system', file=sys.stderr)
        exit(1)
   
    domains = conn.listAllDomains(0)
    if len(domains) != 0:
        for domain in domains:
    	    print(f"Monitoring {domain.name()}")
    	    x = threading.Thread(target=calculate_usage, args=(domain,), daemon=True)
    	    x.start()
    	    threads.append(x)
        y = threading.Thread(target=master, daemon=True)
        y.start()
    else:
        print('  None')
    conn.close()
   
@app.route('/notifications')
def get_notifications():
    global unread
    global read
    return_obj = { "unread":unread.copy(), "read":read.copy() }
    while unread:
        read.append(unread.pop())
    return return_obj
        
@app.route('/stop_monitoring')
def stop_monitoring():
    global started
    started = False
    for t in threads:
        t.join()

@app.route('/')
def hello_world():
    start_monitoring()
    return render_template('index.html')
    
    
    
    
    

